# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["ConcurrencyInterface"]


__doc__ = "A general interface for multi-threading & multi-processing."


from time import sleep

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue

from . import is_async_function
from .loops import asleep, sleep, new_event_loop


class ConcurrencyInterface:
    """
    Defines an interface for managing tasks & pools of threads & processes.
    """

    __slots__ = ()

    _Manager: type

    _default_probe_delay: t.PositiveRealNumber
    _pool: t.PoolExecutorType
    _type: type

    BrokenPool: type

    @classmethod
    def _process_probe_delay(
        cls, value: t.Optional[t.PositiveRealNumber], /
    ) -> t.PositiveRealNumber:
        """
        Ensures the probe frequency is positive & returns it, if it's
        specified. Otherwise, returns the class' default value.
        """
        if value is None:
            return cls._default_probe_delay
        elif value > 0:
            return value
        else:
            raise Issue.value_must("probe_delay", "be > 0")

    @staticmethod
    def _arun_func(
        func: t.Callable[..., t.Any],
        state: t.Sequence[t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Used by the class to retrieve return values from an async or
        sync `func` run in a new process / thread by storing the result
        in a shared `state` container.
        """
        if is_async_function(func):
            run = new_event_loop().run_until_complete
            state.append(run(func(*args, **kwargs)))
        else:
            state.append(func(*args, **kwargs))

    @staticmethod
    def _run_func(
        func: t.Callable[..., t.Any],
        state: t.Sequence[t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Used by the class to retrieve return values from a sync `func`
        run in a new process / thread by storing the result in a shared
        `state` container.
        """
        state.append(func(*args, **kwargs))

    @classmethod
    async def anew(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Runs an async or sync function in another process or thread
        depending on the class calling this method, so that heavy cpu-
        bound computations, or blocking IO operations, can better
        coexist with asynchronous code.
        """
        delay = cls._process_probe_delay(probe_delay)
        state = cls._Manager().list()
        task = cls._type(
            target=cls._arun_func,
            args=(func, state, *args),
            kwargs=kwargs,
        )
        task.start()
        while task.is_alive():
            await asleep(delay)
        task.join()
        return state.pop()

    @classmethod
    def new(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Runs a sync function in another process or thread depending on
        the class calling this method, so that heavy cpu-bound
        computations, or blocking IO operations, can better coexist with
        asynchronous code.
        """
        delay = cls._process_probe_delay(probe_delay)
        state = cls._Manager().list()
        task = cls._type(
            target=cls._run_func,
            args=(func, state, *args),
            kwargs=kwargs,
        )
        task.start()
        while task.is_alive():
            sleep(delay)
        task.join()
        return state.pop()

    @staticmethod
    def _get_result(
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Returns the result of `func` run in a new process / thread using
        the class' pool.
        """
        if is_async_function(func):
            run = new_event_loop().run_until_complete
            return run(func(*args, **kwargs))
        else:
            return func(*args, **kwargs)

    @staticmethod
    def _package_result_methods(
        future: t.Future, /, *, probe_delay: t.PositiveRealNumber
    ) -> t.Future:
        """
        Inserts methods in the `future` returned from a pool submission
        to provide a consistent interface for retrieving results.
        """

        async def aresult() -> t.Any:
            while not future.done():
                await asleep(probe_delay)
            return future._original_result()

        def result() -> t.Any:
            while not future.done():
                sleep(probe_delay)
            return future._original_result()

        future._original_result = future.result
        future.aresult = aresult
        future.result = result
        return future

    @classmethod
    async def asubmit(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Future:
        """
        Submits an async, or synchronous `func` to a process pool or
        thread pool, depending on the class that calls this method, with
        the supplied `*args` & `**kwargs`, then returns the `Future`
        object that's created.
        """
        delay = cls._process_probe_delay(probe_delay)
        future = cls._pool.submit(cls._get_result, func, *args, **kwargs)
        return cls._package_result_methods(future, probe_delay=delay)

    @classmethod
    def submit(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Future:
        """
        Submits a synchronous `func` to a process pool or thread pool,
        depending on the class that calls this method, with the supplied
        `*args` & `**kwargs`, then returns the `Future` object that's
        created.
        """
        delay = cls._process_probe_delay(probe_delay)
        future = cls._pool.submit(func, *args, **kwargs)
        return cls._package_result_methods(future, probe_delay=delay)

    @classmethod
    async def agather(
        cls,
        /,
        *functions: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Mapping[t.Hashable, t.Any] = {},
    ) -> t.List[t.Any]:
        """
        Sumbits all of the async or synchronous `functions` to the
        `Processes._pool` or `Threads._pool` with the given `args` &
        `kwargs`.
        """
        tasks = [
            await cls.asubmit(func, *args, **kwargs) for func in functions
        ]
        try:
            return [await task.aresult() for task in tasks]
        finally:
            for task in tasks:
                task.cancel()

    @classmethod
    def gather(
        cls,
        /,
        *functions: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Mapping[t.Hashable, t.Any] = {},
    ) -> t.List[t.Any]:
        """
        Sumbits all the `functions` to the `Processes._pool` or
        `Threads._pool` with the given `args` & `kwargs`.
        """
        tasks = [cls.submit(func, *args, **kwargs) for func in functions]
        try:
            return [task.result() for task in tasks]
        finally:
            for task in tasks:
                task.cancel()

    @classmethod
    def reset_pool(cls, /) -> None:
        """
        When a process or thread pool is broken by an abruptly exited,
        this method can be called to reset the class' pool object with
        a new instance.
        """
        cls._pool = cls._pool.__class__()


module_api = dict(
    ConcurrencyInterface=t.add_type(ConcurrencyInterface),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_Typing_class import *


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = []


__doc__ = "A collection of (async) generator utilities."


import io
import asyncio
from types import GeneratorType
from types import AsyncGeneratorType
from collections.abc import Iterable, Iterator
from collections.abc import AsyncIterable, AsyncIterator
from inspect import isasyncgenfunction as is_async_gen_function
from inspect import isgeneratorfunction as is_generator_function

from ._typing import Typing as t
from ._constants import BIG
from ._exceptions import Issue, raise_exception


def is_async_iterable(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` supports async iteration.
    """
    return isinstance(obj, AsyncIterable)


def is_iterable(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` supports iteration.
    """
    return isinstance(obj, Iterable)


def is_async_iterator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an async iterator.
    """
    return isinstance(obj, AsyncIterator)


def is_iterator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an iterator.
    """
    return isinstance(obj, Iterator)


def is_async_generator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an async generator.
    """
    return isinstance(obj, AsyncGeneratorType)


def is_generator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an generator.
    """
    return isinstance(obj, GeneratorType)


async def abatch(
    sequence: bytes, /, *, size: int, buffer_type: type = io.BytesIO
) -> t.AsyncGenerator[bytes, None]:
    """
    Runs through a sequence & yields `size` sized chunks of the bytes
    sequence one chunk at a time. By default this async generator yields
    all elements in the sequence.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    sequence = 4 * b" Testing data..."

    async for piece in abatch(sequence, size=32):
        print(piece)
    >>> b' Testing data... Testing data...'
        b' Testing data... Testing data...'

    async for piece in abatch(sequence, size=64):
        print(piece)
    >>> b' Testing data... Testing data... Testing data... Testing data...'
    """
    try:
        read = buffer_type(sequence).read
        while True:
            await asyncio.sleep(0)
            yield read(size) or raise_exception(StopIteration)
    except StopIteration:
        pass


def batch(
    sequence: bytes, /, *, size: int, buffer_type: type = io.BytesIO
) -> t.Generator[bytes, None, None]:
    """
    Runs through a sequence & yields `size` sized chunks of the bytes
    sequence one chunk at a time. By default this generator yields all
    elements in the sequence.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    sequence = 4 * b" Testing data..."

    for piece in batch(sequence, size=32):
        print(piece)
    >>> b' Testing data... Testing data...'
        b' Testing data... Testing data...'

    for piece in batch(sequence, size=64):
        print(piece)
    >>> b' Testing data... Testing data... Testing data... Testing data...'
    """
    try:
        read = buffer_type(sequence).read
        while True:
            yield read(size) or raise_exception(StopIteration)
    except StopIteration:
        pass


async def aresize(
    sequence: t.Sequence[t.Any], /, size: int, *, blocks: int = 0
) -> t.AsyncGenerator[t.Sequence[t.Any], None]:
    """
    Runs through a `sequence` & yields `size` sized chunks of the
    sequence one chunk at a time. `blocks` is the total number of
    chunks allowed to be yielded from the generator. By default this
    generator yields all elements in the sequence.
    """
    length = len(sequence)
    if not blocks or (blocks * size >= length):
        blocks = length + size
    else:
        blocks = (blocks * size) + 1
    async for previous_end, end in azip(
        range(0, blocks, size), range(size, blocks, size)
    ):
        yield sequence[previous_end:end]


def resize(
    sequence: t.Sequence[t.Any], /, size: int, *, blocks: int = 0
) -> t.Generator[t.Sequence[t.Any], None, None]:
    """
    Runs through a `sequence` & yields `size` sized chunks of the
    sequence one chunk at a time. `blocks` is the total number of
    chunks allowed to be yielded from the generator. By default this
    generator yields all elements in the sequence.
    """
    length = len(sequence)
    if not blocks or (blocks * size >= length):
        blocks = length + size
    else:
        blocks = (blocks * size) + 1
    for previous_end, end in zip(
        range(0, blocks, size), range(size, blocks, size)
    ):
        yield sequence[previous_end:end]


async def aunpack(
    iterable: t.AsyncOrSyncIterable[t.Any], /
) -> t.AsyncGenerator[t.Any, None]:
    """
    Runs through an iterable &/or async iterable & yields elements one
    at a time.
    """
    if is_async_iterable(iterable):
        async for item in iterable:
            yield item
    else:
        for item in iterable:
            await asyncio.sleep(0)
            yield item


def unpack(
    iterable: t.Iterable[t.Any], /
) -> t.Generator[t.Any, None, None]:
    """
    Runs through an iterable & yields elements one at a time.
    """
    yield from iterable


async def azip(
    *iterables: t.AsyncOrSyncIterable[t.Any]
) -> t.AsyncGenerator[t.List[t.Any], None]:
    """
    Creates an asynchronous version of the `builtins.zip` function
    which is wrapped by the `Comprende` class.
    """
    coroutines = [aunpack(iterable).__anext__ for iterable in iterables]
    try:
        while True:
            yield [await coroutine() for coroutine in coroutines]
    except (StopAsyncIteration, StopIteration):
        pass


async def acycle(
    iterable: t.AsyncOrSyncIterable[t.Any], /
) -> t.AsyncGenerator[t.Any, None]:
    """
    Unendingly cycles in order over the elements of an async iterable.
    """
    results = []
    if is_async_iterable(iterable):
        async for result in iterable:
            yield result
            results.append(result)
    else:
        for result in iterable:
            await asyncio.sleep(0)
            yield result
            results.append(result)
    while results:
        for result in results:
            await asyncio.sleep(0)
            yield result


def cycle(iterable: t.Iterable[t.Any], /) -> t.Generator[t.Any, None, None]:
    """
    Unendingly cycles in order over the elements of a sync iterable.
    """
    results = []
    for result in iterable:
        yield result
        results.append(result)
    while results:
        for result in results:
            yield result


async def abytes_count(
    *, start: int = 0, size: int = 8, byte_order: str = BIG
) -> t.AsyncGenerator[bytes, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    index = start
    while True:
        await asyncio.sleep(0)
        yield index.to_bytes(size, byte_order)
        index += 1


def bytes_count(
    *, start: int = 0, size: int = 8, byte_order: str = BIG
) -> t.Generator[bytes, None, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    index = start
    while True:
        yield index.to_bytes(size, byte_order)
        index += 1


async def acount(*, start: int = 0) -> t.AsyncGenerator[int, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    if start.__class__ is not int:
        raise Issue.value_must_be_type("start", int)
    index = start
    while True:
        await asyncio.sleep(0)
        yield index
        index += 1


def count(*, start: int = 0) -> t.Generator[int, None, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    if start.__class__ is not int:
        raise Issue.value_must_be_type("start", int)
    index = start
    while True:
        yield index
        index += 1


async def abytes_range(
    *a, size: int = 8, byte_order: str = BIG
) -> t.AsyncGenerator[int, None]:
    """
    An async version of `builtins.range` wrapped by the `Comprende`
    class, & returns its values as bytes instead.
    """
    for result in range(*a):
        await asyncio.sleep(0)
        yield result.to_bytes(size, byte_order)


def bytes_range(
    *a, size: int = 8, byte_order: str = BIG
) -> t.Generator[int, None, None]:
    """
    A synchronous version of `builtins.range` which is wrapped by the
    `Comprende` class, & returns its values as bytes instead.
    """
    for result in range(*a):
        yield result.to_bytes(size, byte_order)


async def arange(*a: t.Optional[int]) -> t.AsyncGenerator[int, None]:
    """
    An async version of `builtins.range`.
    """
    for result in range(*a):
        await asyncio.sleep(0)
        yield result


async def acollate(
    *iterables: t.AsyncOrSyncIterable[t.Any]
) -> t.AsyncGenerator[t.Any, None]:
    """
    Takes a collection of iterables &/or async iterables & exhausts them
    one at a time from left to right.
    """
    for iterable in iterables:
        if is_async_iterable(iterable):
            async for result in iterable:
                yield result
        else:
            for result in iterable:
                await asyncio.sleep(0)
                yield result


def collate(
    *iterables: t.Iterable[t.Any]
) -> t.Generator[t.Any, None, None]:
    """
    Takes a collection of iterables & exhausts them one at a time from
    left to right.
    """
    for iterable in iterables:
        for result in iterable:
            yield result


async def apopleft(
    queue: t.SupportsPopleft, /
) -> t.AsyncGenerator[t.Any, None]:
    """
    An async generator which calls the `popleft()` method on `queue`
    for every iteration, & exits on `IndexError`.
    """
    while True:
        try:
            yield queue.popleft()
        except IndexError:
            break


def popleft(
    queue: t.SupportsPopleft, /
) -> t.Generator[t.Any, None, None]:
    """
    A generator which calls the `popleft()` method on `queue` for
    every iteration, & exits on `IndexError`.
    """
    while True:
        try:
            yield queue.popleft()
        except IndexError:
            break


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    abatch=abatch,
    abytes_count=abytes_count,
    abytes_range=abytes_range,
    acollate=acollate,
    acount=acount,
    acycle=acycle,
    apopleft=apopleft,
    arange=arange,
    aresize=aresize,
    aunpack=aunpack,
    azip=azip,
    batch=batch,
    bytes_count=bytes_count,
    bytes_range=bytes_range,
    collate=collate,
    count=count,
    cycle=cycle,
    is_async_iterable=is_async_iterable,
    is_async_iterator=is_async_iterator,
    is_async_generator=is_async_generator,
    is_async_gen_function=is_async_gen_function,
    is_generator=is_generator,
    is_generator_function=is_generator_function,
    is_iterable=is_iterable,
    is_iterator=is_iterator,
    popleft=popleft,
    resize=resize,
    unpack=unpack,
)




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["AuthFail", "CipherStreamProperties"]


__doc__ = (
    "A definition of shared properties between cipher streaming types."
)


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue, TimestampExpired
from aiootp._exceptions import InvalidBlockID, InvalidSHMAC
from aiootp.commons import OpenFrozenSlots


class AuthFail(OpenFrozenSlots):
    """
    Creates efficient containers for data lost in a buffer during
    authentication failure of a block ID in (Async)DecipherStream
    objects.
    """

    __slots__ = ("block_id", "block", "buffer")

    def __init__(
        self, block_id: bytes, block: bytes, buffer: t.Callable
    ) -> None:
        self.block_id = block_id
        self.block = block
        self.buffer = buffer


class CipherStreamProperties:
    """
    A definition of shared properties between cipher streaming types.
    """

    InvalidBlockID: type = InvalidBlockID
    InvalidSHMAC: type = InvalidSHMAC
    TimestampExpired: type = TimestampExpired

    @property
    def PACKETSIZE(self) -> int:
        """
        Returns the number of combined bytes each iteration of the
        stream will produce. Equal to the header bytes + the blocksize.
        """
        return self._config.PACKETSIZE

    @property
    def aad(self) -> bytes:
        """
        An arbitrary bytes value that a user decides to categorize
        keystreams. It's authenticated as associated data & safely
        differentiates keystreams as a tweak when it's unique for
        each permutation of `key`, `salt`, & `iv`.
        """
        return self._key_bundle.aad

    @property
    def salt(self) -> bytes:
        """
        A [pseudo]random salt that may be supplied by the user.
        By default it's sent in the clear attached to the
        ciphertext. Thus it may simplify implementing efficient
        features, such as search or routing, though care must still
        be taken when considering how leaking such metadata may be
        harmful. Keeping this value constant is strongly discouraged,
        though the salt misuse-reuse resistance of the cipher
        extends up to ~256**(len(iv)/2 + len(siv_key)/2)
        encryptions/second.
        """
        return self._key_bundle.salt

    @property
    def iv(self) -> bytes:
        """
        An ephemeral, uniform, random value that's generated by
        the encryption algorithm. Ensures salt misue / reuse
        security even if the `key`, `salt`, & `aad` are the same for
        ~256**(len(iv)/2 + len(siv_key)/2) encryptions/second.
        """
        return self._key_bundle.iv


module_api = dict(
    AuthFail=t.add_type(AuthFail),
    CipherStreamProperties=t.add_type(CipherStreamProperties),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import platform

from test_initialization import *

from aiootp._paths import DatabasePath
from aiootp.asynchs import sleep
from aiootp.keygens.passcrypt.config import passcrypt_spec


def test_sign_and_verify():
    try:
        test_key = token_bytes(64)
        test_salt = token_bytes(64)
        test_directory = DatabasePath()

        signer = PackageSigner(
            package=aiootp.__package__,
            version=aiootp.__version__,
            author=aiootp.__author__,
            license=aiootp.__license__,
            description=aiootp.__doc__,
        )
        signer.update_scope(build_number=0)
        _repr = repr(signer)
        assert all(name in _repr for name in signer._scope)

        problem = (
            "Allowed to retrieve a signature before connecting to the "
            "database."
        )
        with Ignore(RuntimeError, if_else=violation(problem)):
            signer._signature

        is_mac_os_issue = lambda relay: (platform.system() == "Darwin")
        while True:
            sleep(0.001)
            with Ignore(ConnectionRefusedError, if_except=is_mac_os_issue):
                signer.connect_to_secure_database(
                    username=b"test_username",
                    passphrase=test_key,
                    salt=test_salt,
                    path=test_directory,
                    mb=passcrypt_spec.MIN_MB,
                    cpu=1,
                )
                break
        signer.update_public_credentials(x25519_public_key=aiootp.__PUBLIC_X25519_KEY__)

        problem = (
            "Allowed to retrieve a signing key before its creation."
        )
        with Ignore(LookupError, if_else=violation(problem)):
            signer.signing_key

        problem = (
            "Allowed to retrieve a signature before signing."
        )
        with Ignore(RuntimeError, if_else=violation(problem)):
            signer._signature

        test_signing_key = signer.generate_signing_key()

        # update_signing_key expects a str, bytes, Ed25519PrivateKey or an Ed25519 object
        signer.update_signing_key(test_signing_key)
        signer.update_signing_key(signer.signing_key.secret_key)
        signer.update_signing_key(signer.signing_key.secret_bytes)

        problem = (
            "A type other than str, bytes, Ed25519PrivateKey or an "
            "Ed25519 object was allowed for update_signing_key."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            signer.update_signing_key(bytearray(signer.signing_key.secret_bytes))

        filename_sheet = """
        include tests/test_initialization.py
        include tests/test_aiootp.py
        include tests/test_generics.py
        include tests/test_ByteIO.py
        include tests/test_misc_in_generics.py
        include tests/test_randoms.py
        include tests/test_misc_in_randoms.py
        include tests/test_ciphers.py
        include tests/test_Database_AsyncDatabase.py
        include tests/test_StreamHMAC.py
        include tests/test_X25519_Ed25519.py
        include tests/test_high_level_encryption.py
        include tests/test_misc_in_ciphers.py
        include tests/test_Passcrypt.py
        """.strip().split("\n")

        test_path = Path(__file__).parent.parent
        for line in filename_sheet:
            filename = line.strip().split(" ")[-1]
            with open(test_path / filename, "rb") as source_file:
                file_data = source_file.read()
                signer.add_file(filename, file_data)

                # added file hashes are available from the files attribute
                # by their filename
                assert filename in signer.files
                assert signer.files[filename].digest() != signer._Hasher().digest()
                assert signer.files[filename].digest() == signer._Hasher(file_data).digest()

        signer.sign_package()
        summary = signer.summarize()

        # The package verifier successfully verifies a correct summary
        verifier = PackageVerifier(signer.signing_key.public_bytes, path=test_path)
        verifier.verify_summary(summary)

        # Package verifier can also accept an Ed25519 object
        verifier = PackageVerifier(
            Ed25519().import_public_key(signer.signing_key.public_key), path=test_path
        )
        verifier.verify_summary(summary)

        # summarize calls produce the same summaries if the signer object
        # has not been modified
        assert summary == signer.summarize()

        # summary checksum alteration fails
        summary["checksum"] = summary["checksum"][::-1]
        problem = (
            "Summary alteration uncaught."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            verifier.verify_summary(summary)

        # returning the checksum to the original value succeeds
        summary["checksum"] = summary["checksum"][::-1]
        verifier.verify_summary(summary)

        # a json summary also works
        verifier.verify_summary(json.dumps(summary))

        # altering the signature does not work
        problem = (
            "Signature alteration went uncaught."
        )
        with Ignore(verifier.InvalidSignature, if_else=violation(problem)):
            fake_summary = {**summary, "signature": token_bytes(64).hex()}
            verifier.verify_summary(fake_summary)

        # the signing key in the summary is the hex representation of the
        # signing objects public key
        assert len(summary["signing_key"]) == 64
        assert summary["signing_key"] == signer.signing_key.public_bytes.hex()

        # altering the signing key fails
        summary["signing_key"] = X25519().generate().public_bytes.hex()
        problem = (
            "Changed signing_key went uncaught."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            verifier.verify_summary(summary)

        # returning to the original signing key succeeds
        summary["signing_key"] = test_signing_key.public_bytes.hex()
        verifier.verify_summary(summary)

        # changing the signature is detected during summarization
        VERSION = signer._scope.version
        VERSIONS = signer._VERSIONS
        PACKAGE = signer._scope.package
        signature = signer.db[PACKAGE][VERSIONS][VERSION]
        signer.db[PACKAGE][VERSIONS][VERSION] = token_bytes(32).hex()
        problem = (
            "Altered signature not detected during summarization."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            signer.summarize()

        # returning the signature works
        signer.db[PACKAGE][VERSIONS][VERSION] = signature
        signer.summarize()

        # verifier detects wrong file digest
        signer.sign_package()
        summary = signer.summarize()
        filename = list(summary[CHECKSUMS])[-1]
        summary[CHECKSUMS][filename] = signer._Hasher().hexdigest()
        problem = (
            "An invalid file digest wasn't detected."
        )
        with Ignore(InvalidDigest, if_else=violation(problem)):
            verifier._verify_file_checksums(summary)

        # verifier throws error if path is specified without also
        # specifying direct file validation.
        problem = (
            "A verifier was initialized with conflicting flags."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            PackageVerifier(
                signer.signing_key.public_bytes,
                path=test_path,
                verify_files=False,
            )
    finally:
        signer.db.delete_database()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from random import randrange

from test_initialization import *


class TestOnlineCipherInterfaces:

    async def test_sync_cipher_decipher_streams_with_varied_data_sizes(
        self
    ) -> None:
        for (config, cipher, salt, aad) in all_ciphers:
            for i in range(0, 512 - config.INNER_HEADER_BYTES - 1, randrange(1, 64)):
                stream_enc = cipher.stream_encrypt(salt=salt, aad=aad)
                pt_enc = i * b"a"
                stream_enc.buffer(pt_enc)

                stream_dec = cipher.stream_decrypt(
                    salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
                )
                for var in ("salt", "iv", "aad", "PACKETSIZE"):
                    assert getattr(stream_enc, var) == getattr(stream_dec, var)
                pt_dec = b""
                join = b"".join
                for id_ct in stream_enc:
                    problem = (
                        "Invalid packet sizes were allowed."
                    )
                    with Ignore(ValueError, if_else=violation(problem)):
                        stream_dec.buffer(join(id_ct)[:stream_dec.PACKETSIZE - 1])

                    stream_dec.buffer(join(id_ct))
                    for pt in stream_dec:
                        pt_dec += pt
                for id_ct in stream_enc.finalize():
                    stream_dec.buffer(join(id_ct))
                for pt in stream_dec.finalize():
                    pt_dec += pt

                problem = (
                    "Processing was allowed to continue after finalization."
                )
                with Ignore(InterruptedError, if_else=violation(problem)):
                    stream_enc.buffer(pt_enc)
                with Ignore(InterruptedError, if_else=violation(problem)):
                    stream_dec.buffer(id_ct[1])
                assert pt_enc == pt_dec, f"{i=} : plaintext_len={len(pt_enc)} : decrypted_plaintext_len={len(pt_dec)}"

    async def test_async_cipher_decipher_streams_with_varied_data_sizes(
        self
    ) -> None:
        for (config, cipher, salt, aad) in all_ciphers:
            for i in range(0, 512 - config.INNER_HEADER_BYTES - 1, randrange(1, 64)):
                stream_enc = await cipher.astream_encrypt(salt=salt, aad=aad)
                pt_enc = i * b"a"
                await stream_enc.abuffer(pt_enc)

                stream_dec = await cipher.astream_decrypt(
                    salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
                )
                for var in ("salt", "iv", "aad", "PACKETSIZE"):
                    assert getattr(stream_enc, var) == getattr(stream_dec, var)
                pt_dec = b""
                join = b"".join
                async for id_ct in stream_enc:
                    problem = (
                        "Invalid packet sizes were allowed."
                    )
                    with Ignore(ValueError, if_else=violation(problem)):
                        await stream_dec.abuffer(join(id_ct)[:stream_dec.PACKETSIZE - 1])

                    await stream_dec.abuffer(join(id_ct))
                    async for pt in stream_dec:
                        pt_dec += pt
                async for id_ct in stream_enc.afinalize():
                    await stream_dec.abuffer(join(id_ct))
                async for pt in stream_dec.afinalize():
                    pt_dec += pt

                problem = (
                    "Processing was allowed to continue after finalization."
                )
                with Ignore(InterruptedError, if_else=violation(problem)):
                    await stream_enc.abuffer(pt_enc)
                with Ignore(InterruptedError, if_else=violation(problem)):
                    await stream_dec.abuffer(id_ct[1])
                assert pt_enc == pt_dec, f"{i=} : plaintext_len={len(pt_enc)} : decrypted_plaintext_len={len(pt_dec)}"

    async def test_async_encipher_sync_decipher_interop(
        self
    ) -> None:
        for (config, cipher, salt, aad) in all_ciphers:
            stream_enc = await cipher.astream_encrypt(salt=salt, aad=aad)
            pt_enc = plaintext_bytes
            await stream_enc.abuffer(pt_enc)

            stream_dec = cipher.stream_decrypt(
                salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
            )
            for var in ("salt", "iv", "aad", "PACKETSIZE"):
                assert getattr(stream_enc, var) == getattr(stream_dec, var)
            pt_dec = b""
            join = b"".join

            async for block_id, block in stream_enc:
                # the cipher can be resumed after the failure state of an
                # InvalidBlockID exception is recovered & the correct data
                # is supplied to the buffer
                problem = (
                    "An altered block_id was not detected."
                )
                fake_block_id = xi_mix(block_id + b"\x01", size=len(block_id))
                with Ignore(InvalidBlockID, if_else=violation(problem)) as relay:
                    stream_dec.buffer(fake_block_id + block)
                assert fake_block_id == relay.error.failure_state.block_id
                assert block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                problem = (
                    "An altered block was not detected."
                )
                fake_block = xi_mix(block + b"\x01", size=len(block))
                with Ignore(InvalidBlockID, if_else=violation(problem)) as relay:
                    stream_dec.buffer(block_id + fake_block)
                assert block_id == relay.error.failure_state.block_id
                assert fake_block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                # correct data is processed without failure
                stream_dec.buffer(block_id + block)
                for pt in stream_dec:
                    pt_dec += pt
            async for id_ct in stream_enc.afinalize():
                stream_dec.buffer(join(id_ct))
            for pt in stream_dec.finalize():
                pt_dec += pt
            assert pt_dec == pt_enc

    async def test_sync_encipher_async_decipher_interop(
        self
    ) -> None:
        for (config, cipher, salt, aad) in all_ciphers:
            stream_enc = cipher.stream_encrypt(salt=salt, aad=aad)
            pt_enc = plaintext_bytes
            stream_enc.buffer(pt_enc)

            stream_dec = await cipher.astream_decrypt(
                salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
            )
            for var in ("salt", "iv", "aad", "PACKETSIZE"):
                assert getattr(stream_enc, var) == getattr(stream_dec, var)
            pt_dec = b""
            join = b"".join

            for block_id, block in stream_enc:
                # the cipher can be resumed after the failure state of an
                # InvalidBlockID exception is recovered & the correct data
                # is supplied to the buffer
                problem = (
                    "An altered block_id was not detected."
                )
                fake_block_id = xi_mix(block_id + b"\x01", size=len(block_id))
                async with Ignore(InvalidBlockID, if_else=violation(problem)) as relay:
                    await stream_dec.abuffer(fake_block_id + block)
                assert fake_block_id == relay.error.failure_state.block_id
                assert block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                problem = (
                    "An altered block was not detected."
                )
                fake_block = xi_mix(block + b"\x01", size=len(block))
                async with Ignore(InvalidBlockID, if_else=violation(problem)) as relay:
                    await stream_dec.abuffer(block_id + fake_block)
                assert block_id == relay.error.failure_state.block_id
                assert fake_block == relay.error.failure_state.block
                assert 0 == len(relay.error.failure_state.buffer())

                # correct data is processed without failure
                await stream_dec.abuffer(block_id + block)
                async for pt in stream_dec:
                    pt_dec += pt

            for id_ct in stream_enc.finalize():
                await stream_dec.abuffer(join(id_ct))
            async for pt in stream_dec.afinalize():
                pt_dec += pt
            assert pt_dec == pt_enc


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *

from aiootp.commons.instances import *
from aiootp.commons.slots import *
from aiootp.commons.namespaces import *
from aiootp.commons.configs import *


class SlotsAttributes:
    __slots__ = ()

    _items: t.Dict[str, t.Any] = dict(
        _private=True, one=1, mapped="value", unmapped="attr"
    )
    _MAPPED_ATTRIBUTES: t.Tuple[str] = tuple(
        name for name in _items if name != "unmapped"
    )
    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        *Slots._UNMAPPED_ATTRIBUTES, "mapped", "unmapped"
    )                       # Being mapped overrides being unmapped


class NamespaceAttributes(SlotsAttributes):
    __slots__ = ()

    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        *Namespace._UNMAPPED_ATTRIBUTES, "mapped", "unmapped"
    )                       # Being mapped overrides being unmapped


class BaseVariableHoldingClassTests:
    _items = dict(SlotsAttributes._items)
    _type: type
    _open: bool
    _frozen: bool

    async def test_init_accepts_mapping_and_keywords(self) -> None:
        obj_from_mapping = self._type(self._items)
        obj_from_keywords = self._type(**self._items)
        for name, value in self._items.items():
            assert value == getattr(obj_from_mapping, name)
            assert value == getattr(obj_from_keywords, name)


class BaseReprControlledTests(BaseVariableHoldingClassTests):

    async def test_instance_repr_shows_values(self) -> None:
        DebugControl.disable_debugging()
        obj = self._type(self._items)
        string = repr(obj)
        for name, value in self._items.items():
            if name[0] == "_":
                assert name not in string
            elif (
                hasattr(obj, "_is_mapped_attribute")
                and not obj._is_mapped_attribute(name)
            ):
                assert name not in string
            else:
                assert name in string, name
                assert self._open ^ (str(value) not in string), value
                assert self._open ^ (str(value.__class__) in string), value


class BaseMaskableReprTests(BaseVariableHoldingClassTests):

    async def test_mask_hides_or_shows_repr(self) -> None:
        DebugControl.disable_debugging()
        obj = self._type(self._items)
        for name in self._items:
            if (
                str(name) in getattr(obj, "_UNMAPPED_ATTRIBUTES", ())
                or str(name).startswith("_")
            ):
                continue
            assert str(name) in obj.__repr__(mask=False), name
            assert str(getattr(obj, name)) in obj.__repr__(mask=False), name
            assert str(name) in obj.__repr__(mask=True), name
            assert str(getattr(obj, name)) not in obj.__repr__(mask=True), name
            DebugControl.enable_debugging(silence_warnings=True)
            assert str(name) in repr(obj), name
            assert str(getattr(obj, name)) in repr(obj), name
            DebugControl.disable_debugging()


class BaseFrozenTests(BaseVariableHoldingClassTests):

    async def test_instance_values_cant_be_changed_once_set(self) -> None:
        problem = (
            "An instance was allowed to be mutated."
        )
        obj = self._type()
        for name, value in self._items.items():
            setattr(obj, name, value)
            with Ignore(PermissionError, if_else=violation(problem)):
                setattr(obj, name, value)
            with Ignore(PermissionError, if_else=violation(problem)):
                delattr(obj, name)


class BaseModuleNamespaceTests(BaseVariableHoldingClassTests):

    async def test_all_doesnt_include_private_variables(self) -> None:
        obj = self._type(self._items)
        private_name = "_private"
        assert hasattr(obj, private_name)
        assert private_name not in obj.__all__


class BaseDictLikeTests(BaseVariableHoldingClassTests):

    def frozen_violation_catcher(self, name: str) -> Ignore:
        if_except = lambda _: self._frozen
        if_else = lambda _: (
            (not self._frozen) or raise_exception(AssertionError(name))
        )
        return Ignore(PermissionError, if_except=if_except, if_else=if_else)

    async def test_inclusion_exclusion_logic(self) -> None:
        items = {**self._items, 123: "number"}
        obj = self._type()
        unmapped = set()
        MAPPED = getattr(obj, "_MAPPED_ATTRIBUTES", ())
        UNMAPPED = getattr(obj, "_UNMAPPED_ATTRIBUTES", ())
        for i, (name, value) in enumerate(items.items(), start=1):
            if name in MAPPED or name not in UNMAPPED:
                pass
            else:
                unmapped.add(name)
            obj[name] = value
            assert name in obj
            assert value == obj[name]
            assert i == len(obj) + len(unmapped)
            with self.frozen_violation_catcher(name):
                del obj[name]
                assert name not in obj
            with self.frozen_violation_catcher(name):
                obj[name] = value

    async def test_mapping_methods(self) -> None:
        items = {**self._items, 123: "number"}
        obj = self._type()
        for name, value in items.items():
            obj[name] = value

        assert set(items.keys()).issuperset(obj.keys())
        assert set(items.values()).issuperset(obj.values())
        assert set(items.items()).issuperset(obj.items())

        items.update(dict(new_value=True))
        obj.update(dict(new_value=True))
        assert obj["new_value"] == True
        assert obj["new_value"] == items["new_value"]

        with self.frozen_violation_catcher("new_value"):
            obj.update(dict(new_value=True).items())
            assert obj["new_value"] == True
        with self.frozen_violation_catcher("new_value"):
            obj.update(**dict(new_value=False))
            assert obj["new_value"] == False
        with self.frozen_violation_catcher("new_value"):
            obj.update(dict(new_value=False), **dict(new_value=True))
            assert obj["new_value"] == True


class BaseIndexableTests(BaseVariableHoldingClassTests):

    async def test_unmapped_attributes_arent_in_dir(self) -> None:
        obj = self._type(self._items)
        if not hasattr(obj, "_UNMAPPED_ATTRIBUTES"):
            return
        if all((type(item) is str) for item in self._items):
            assert not (
                set(obj.__class__._UNMAPPED_ATTRIBUTES)
                .difference(obj.__class__._MAPPED_ATTRIBUTES)
            ).intersection(value for value in dir(obj))

    async def test_len_is_number_of_mapped_items_in_instance(self) -> None:
        obj = self._type(self._items)
        assert len(obj) == sum(1 for name in obj)
        assert len(obj) == len(
            set(self._items)
            .difference(getattr(obj, "_UNMAPPED_ATTRIBUTES", ()))
            .union(getattr(obj, "_MAPPED_ATTRIBUTES", ()))
        )

    async def test_indexable_iterations(self) -> None:
        obj = self._type(self._items)
        async for name in obj:
            assert name in obj
            assert name in self._items
            assert obj[name] == self._items[name]
        for name in obj:
            assert name in obj
            assert name in self._items
            assert obj[name] == self._items[name]

    async def test_item_indexing(self) -> None:
        names, values = list(self._items), list(self._items.values())
        items = {}
        all_items = {}
        obj = self._type()

        for name, value in zip(names, values):
            obj[name] = value
            all_items[name] = value
            if name in dict(obj):
                items[name] = value
                assert obj[name] == items[name]
            else:
                assert name in obj._UNMAPPED_ATTRIBUTES
            assert obj[name] == all_items[name]

        with Ignore(PermissionError, if_except=lambda _: self._frozen):
            del obj[names[0]]
            del items[names[0]]
            assert dict(obj) == dict(items)

    async def test_false_contains_logic(self) -> None:
        obj = self._type()
        assert randoms.token_bytes(4).hex() not in obj
        assert randoms.token_bits(32) not in obj
        assert None not in obj


# Slots

class SlotsType(SlotsAttributes, Slots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class OpenSlotsType(SlotsAttributes, OpenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenSlotsType(SlotsAttributes, FrozenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenSlotsDictType(NamespaceAttributes, FrozenSlots):
    __slots__ = ("__dict__", *BaseVariableHoldingClassTests._items)


class OpenFrozenSlotsType(SlotsAttributes, OpenFrozenSlots):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class FrozenInstanceType(SlotsAttributes, FrozenInstance):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)


class TestSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = SlotsType
    _open: bool = False
    _frozen: bool = False


class TestOpenSlots(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = OpenSlotsType
    _open: bool = True
    _frozen: bool = False


class TestFrozenSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = FrozenSlotsType
    _open: bool = False
    _frozen: bool = True


class TestFrozenSlotsDict(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = FrozenSlotsDictType
    _open: bool = False
    _frozen: bool = True


class TestOpenFrozenSlots(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseIndexableTests,
):
    _type: type = OpenFrozenSlotsType
    _open: bool = True
    _frozen: bool = True


class TestFrozenInstance(
    BaseReprControlledTests,
    BaseFrozenTests,
):
    _type: type = FrozenInstanceType
    _open: bool = False
    _frozen: bool = True


# Configs

class ConfigType(SlotsAttributes, Config):
    __slots__ = tuple(BaseVariableHoldingClassTests._items)

    slots_types = dict({
        name: BaseVariableHoldingClassTests._items[name].__class__
        for name in BaseVariableHoldingClassTests._items
    })


class TestConfig(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseIndexableTests,
):
    _type: type = ConfigType
    _open: bool = True
    _frozen: bool = True


# Namespaces

class NamespaceType(NamespaceAttributes, Namespace):
    pass


class OpenNamespaceType(NamespaceAttributes, OpenNamespace):
    pass


class FrozenNamespaceType(NamespaceAttributes, FrozenNamespace):
    pass


class OpenFrozenNamespaceType(NamespaceAttributes, OpenFrozenNamespace):
    pass


class TestNamespaceMapping(
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = t.NamespaceMapping
    _open: bool = True
    _frozen: bool = False


class TestNamespace(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = NamespaceType
    _open: bool = False
    _frozen: bool = False


class TestOpenNamespace(
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = OpenNamespaceType
    _open: bool = True
    _frozen: bool = False


class TestFrozenNamespace(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseModuleNamespaceTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = FrozenNamespaceType
    _open: bool = False
    _frozen: bool = True


class TestOpenFrozenNamespace(
    BaseFrozenTests,
    BaseReprControlledTests,
    BaseMaskableReprTests,
    BaseModuleNamespaceTests,
    BaseDictLikeTests,
    BaseIndexableTests,
):
    _type: type = OpenFrozenNamespaceType
    _open: bool = True
    _frozen: bool = True


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "FrozenNamespace", "Namespace", "OpenFrozenNamespace", "OpenNamespace"
]


__doc__ = "Definitions for mapping classes."


import asyncio

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue

from .slots import Slots


class Namespace(Slots):
    """
    A simple wrapper for turning mappings into Namespace objects that
    allow dotted lookup and assignment on those mappings.
    """

    __slots__ = ("__dict__",)

    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        "__dict__", *Slots._UNMAPPED_ATTRIBUTES
    )

    def __init__(
        self, mapping: t.Mapping[t.Hashable, t.Any] = {}, /, **kw: t.Any
    ) -> None:
        """
        Maps the user-defined mapping & kwargs to the Namespace's
        instance dictionary.
        """
        self.__dict__.update(mapping) if mapping else 0
        self.__dict__.update(kw) if kw else 0


class OpenNamespace(Namespace):
    """
    A version of the `Namespace` class which doesn't omit instance
    repr's by default.
    """

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


class FrozenNamespace(Namespace):
    """
    A simple wrapper for turning mappings into Namespace objects that
    allow dotted lookup and assignment on those mappings. Also, provides
    utilities for copying mappings into other containers, & turning
    mappings into stand-alone, first-class modules.
    """

    @property
    def __all__(self, /) -> t.List[str]:
        """
        Allows users that have turned their namespace into a Module
        object to do a `from namespace import *` on the contents of
        the namespace's mapping. This method excludes exporting private
        methods & attributes.
        """
        return [var for var in self if str(var)[0] != "_"]

    def __setattr__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if name in self:
            raise Issue.cant_reassign_attribute(name)
        object.__setattr__(self, name, value)

    def __delattr__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if name in self:
            raise Issue.cant_reassign_attribute(name)
        elif name.__class__ is str:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value

    def __delitem__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)


class OpenFrozenNamespace(FrozenNamespace):
    """
    A version of the `FrozenNamespace` class which doesn't omit instance
    repr's by default.
    """

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


module_api = dict(
    FrozenNamespace=t.add_type(FrozenNamespace),
    Namespace=t.add_type(Namespace),
    OpenFrozenNamespace=t.add_type(OpenFrozenNamespace),
    OpenNamespace=t.add_type(OpenNamespace),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)




# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import json
from getpass import getpass
from setuptools import setup, find_packages

from aiootp import (
    __doc__,
    __license__,
    __package__,
    __version__,
    __author__,
)
from aiootp import PackageSigner, PackageVerifier


with open("README.rst", "r") as readme:
    long_description = readme.read()


if __name__ != "__main__":
    pass
elif getpass("sign package? (y/N) ").lower().strip().startswith("y"):

    with open("SIGNATURE.txt", "r") as sig:
        scope = json.loads(sig.read())[PackageSigner._SCOPE]
        print("current version:", __version__)
        print(f"current build: {scope['build_number']}\n")

    signer = PackageSigner(
        package=__package__,
        version=__version__,
        author=__author__,
        license=__license__,
        description=__doc__,
        build_number=int(getpass("build number: ")),
    )
    signer.connect_to_secure_database(
        username=getpass("database username: ").encode(),
        passphrase=getpass("database key: ").encode(),
        salt=getpass("database salt: ").encode(),
        path=getpass("secure directory: "),
    )

    if getpass("is the signing key already saved on this device? (Y/n) ").lower().strip().startswith("n"):
        signer.update_signing_key(bytes.fromhex(getpass("signing key: ").strip()))

    while getpass("update public credentials? (y/N) ").lower().strip().startswith("y"):
        signer.update_public_credentials(
            **{getpass("name: ").strip(): getpass("value: ")}
        )

    with open("MANIFEST.in", "r") as manifest:
        filename_sheet = manifest.read().strip().split("\n")

    for line in filename_sheet:
        line = line.strip()
        if "SIGNATURE" in line or not line.startswith("include"):
            continue
        filename = line.split(" ")[-1]
        with open(filename, "rb") as source_file:
            signer.add_file(filename, source_file.read())

    signer.sign_package()
    signer.db.save_database()
    summary = signer.summarize()
    verifier = PackageVerifier(signer.signing_key.public_bytes, path="")
    verifier.verify_summary(summary)
    with open("SIGNATURE.txt", "w+") as attestation:
        attestation.write(json.dumps(summary, indent=4))


setup(
    name=__package__,
    license=__license__,
    version=__version__,
    description=__doc__,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://twitter.com/aiootp",
    author="Gonzo Investigative Journalism Agency, LLC",
    author_email="gonzo.development@protonmail.ch",
    maintainer="Gonzo Investigative Journalism Agency, LLC",
    maintainer_email="gonzo.development@protonmail.ch",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.8",
    tests_require=["pytest>=8.1.1", "pytest-asyncio>=0.23.6"],
    install_requires=["aiofiles>=23.2.1", "cryptography>=42.0.8"],
    options={"bdist_wheel": {"python_tag": "py38.py39.py310.py311.py312"}},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Framework :: IPython",
        "Framework :: Jupyter",
        "Framework :: Pytest",
        "Natural Language :: English",
        "Operating System :: Other OS",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Operating System :: OS Independent",
        "Operating System :: Microsoft :: Windows",
        "Topic :: System",
        "Topic :: Internet",
        "Topic :: Security",
        "Topic :: Database",
        "Topic :: Education",
        "Topic :: Utilities",
        "Topic :: Communications",
        "Topic :: Office/Business",
        "Topic :: Text Processing",
        "Topic :: System :: Archiving",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development",
        "Topic :: System :: Networking",
        "Topic :: System :: Filesystems",
        "Topic :: Communications :: Chat",
        "Topic :: Scientific/Engineering",
        "Topic :: Communications :: Email",
        "Topic :: Security :: Cryptography",
        "Topic :: Office/Business :: Financial",
        "Topic :: Communications :: File Sharing",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Software Distribution",
        "Topic :: System :: Systems Administration",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Office/Business :: Financial :: Accounting",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Typing :: Typed",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Intended Audience :: Legal Industry",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Telecommunications Industry",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ],
    keywords=" ".join(
        [
            "xor key salt pepper nonce aad iv siv resuse misuse",
            "fully context committing commitment",
            "online AEAD auth authenticated authentication",
            "shmac hmac nmac mac digest integrity",
            "infosec opsec appsec privacy engineering",
            "stream block cipher permutation chunky2048 slick256",
            "encrypt plaintext",
            "decrypt ciphertext",
            "passcrypt passphrase PBKDF",
            "password based derivation function",
            "3dh 2dh 25519 x25519 ed25519 curve25519",
            "diffie hellman sign signature verify verification",
            "db database value store",
            "user uuid unique guid global",
            "transparent encryption decryption",
            "indistinguishable indistinguishability",
            "pseudo one time pad onetimepad",
            "canonical canonicalization domain separation KDF",
            "bit bits 64 128 256 512 1024 2048 4096",
            "hash sha sha3 sha-3 keccak",
            "ephemeral byte entropy",
            "PRF PRG PRP RNG PRNG CSPRNG",
            "cryptographically secure",
            "random number generator",
            "bitwise operations",
            "IND CCA CPA RUP",
            "information cyber security",
            "active passive adaptive chosen attack",
            "resistance resistant tweak tweakable",
            "anonymous anonymity pseudonymous",
            "symmetric asymmetric",
            "communications utilities",
            "simple clean code",
            "crypto cryptology cryptography cryptanalysis",
            "beta testing",
            "data science multi processing threading",
            "await async asyncio parallel concurrency",
            "coroutine coroutines asynchronous asynchrony",
        ]
    ),
) if __name__ == "__main__" else 0

